/*
 *  copyright 2003
 *
 *  MrBayes 3.0
 *
 *  John P. Huelsenbeck
 *  Section of Ecology, Behavior and Evolution
 *  Division of Biological Sciences
 *  University of California, San Diego
 *  La Jolla, CA 92093-0116
 *
 *  johnh@biomail.ucsd.edu
 *
 *	Fredrik Ronquist
 *  Dept. Systematic Zoology
 *  Evolutionary Biology Centre
 *	Uppsala University
 *	Norbyv. 18D, SE-752 36 Uppsala, Sweden
 *
 *  fredrik.ronquist@ebc.uu.se
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "mb.h"
#include "globals.h"
#include "command.h"
#include "bayes.h"
#include "sump.h"
#if defined(__MWERKS__)
#include "SIOUX.h"
#endif
/* local prototypes */
void     GetSumpToken (int *tokenType);
int      MeanVariance(int col, MrBFlt *m, MrBFlt *v, int isLog);
void     Sort2 (MrBFlt *item, int left, int right);
int      SortParameters (void);
/* local (to this file) */
int			numColumns, numRows;
MrBFlt		*parameterValues;
char		*sumpTokenP, sumpToken[CMD_STRING_LENGTH], *headerNames;
int DoSump (void)
{
	int			i, j, k, lineTerm, longestLineLength, tokenType, lineNum, lastTokenWasDash,
				inSumpComment, allDigitLine, nNumbersOnThisLine, lastNonDigitLine,
				numParamLines, numLinesToRead, numLinesRead, firstNumCols, nLines,
				nHeaders, len, longestHeader, whichIsX, whichIsY, screenWidth, screenHeigth, numY[60],
				aaModelCounts[20], startReading;
	MrBFlt		tempD, mean, var, lower, upper, minX, minY, maxX, maxY,
				meanY[60], x, y;
	char		*s, *headerLine, temp[100], val[100], parts[100];
	FILE		*fp;
	
#	if defined (MPI_ENABLED)
	int			ierror;
#	endif
	
#	if defined (MPI_ENABLED)
	if (proc_id == 0)
		{
#	endif
	/* set file pointer to NULL */
	fp = NULL;
	/* Check that a data set has been read in. We check taxon names against
	   those read in. */
	if (defMatrix == NO)
		{
		MrBayesPrint ("%s   A matrix must be specified before sump can be used\n", spacer);
		goto errorExit;
		}
	/* tell user we are ready to go */
	MrBayesPrint ("%s   Summarizing parameters in file %s\n", spacer, sumpParams.sumpFileName);
	
	/* open binary file */
	if ((fp = OpenBinaryFileR(sumpParams.sumpFileName)) == NULL)
		goto errorExit;
		
	/* find out what type of line termination is used */
	lineTerm = LineTermType (fp);
	if (lineTerm == LINETERM_MAC)
		MrBayesPrint ("%s   Macintosh line termination\n", spacer);
	else if (lineTerm == LINETERM_DOS)
		MrBayesPrint ("%s   DOS line termination\n", spacer);
	else if (lineTerm == LINETERM_UNIX)
		MrBayesPrint ("%s   UNIX line termination\n", spacer);
	else
		{
		MrBayesPrint ("%s   Unknown line termination\n", spacer);
		goto errorExit;
		}
		
	/* find length of longest line */
	longestLineLength = LongestLine (fp);
	MrBayesPrint ("%s   Longest line length = %d\n", spacer, longestLineLength);
	longestLineLength += 10;
	
	/* allocate a string long enough to hold a line */
	if (memAllocs[ALLOC_SUMPSTRING] == YES)
		{
		MrBayesPrint ("%s   Sump string is already allocated\n", spacer);
		goto errorExit;
		}
	s = (char *)malloc((size_t) (longestLineLength * sizeof(char)));
	if (!s)
		{
		MrBayesPrint ("%s   Problem allocating string for reading sump file\n", spacer);
		goto errorExit;
		}
	headerLine = (char *)malloc((size_t) (longestLineLength * sizeof(char)));
	if (!headerLine)
		{
		MrBayesPrint ("%s   Problem allocating headerLine for reading sump file\n", spacer);
		goto errorExit;
		}
	headerNames = (char *)malloc((size_t) ((longestLineLength+40) * sizeof(char)));
	if (!headerNames)
		{
		MrBayesPrint ("%s   Problem allocating headerNames for reading sump file\n", spacer);
		goto errorExit;
		}
	for (i=0; i<longestLineLength+40; i++)
		headerNames[i] = ' ';
	headerNames[longestLineLength+40-1] = '\0';
	memAllocs[ALLOC_SUMPSTRING] = YES;
		
	/* close binary file */
	fclose (fp);
	
	/* open text file */
	if ((fp = OpenTextFileR(sumpParams.sumpFileName)) == NULL)
		goto errorExit;
	
	/* Check file for appropriate blocks. We want to find the last block
	   in the file and start from there. */
	inSumpComment = NO;
	lineNum = lastNonDigitLine = numParamLines = 0;
	while (fgets (s, longestLineLength, fp) != NULL)
		{
		sumpTokenP = &s[0];
		allDigitLine = YES;
		lastTokenWasDash = NO;
		nNumbersOnThisLine = 0;
		do
			{
			GetSumpToken (&tokenType);
			/*printf ("%s (%d)\n", sumpToken, tokenType);*/
			if (IsSame("[", sumpToken) == SAME)
				inSumpComment = YES;
			if (IsSame("]", sumpToken) == SAME)
				inSumpComment = NO;
				
			if (inSumpComment == NO)
				{
				if (tokenType == NUMBER)
					{
					sscanf (sumpToken, "%lf", &tempD);
					if (lastTokenWasDash == YES)
						tempD *= -1.0;
					nNumbersOnThisLine++;
					lastTokenWasDash = NO;
					}
				else if (tokenType == DASH)
					{
					lastTokenWasDash = YES;
					}
				else if (tokenType != UNKNOWN_TOKEN_TYPE)
					{
					allDigitLine = NO;
					lastTokenWasDash = NO;
					}
				
				}
				
			} while (*sumpToken);
		lineNum++;
		
		if (allDigitLine == NO)
			{
			lastNonDigitLine = lineNum;
			numParamLines = 0;
			strcpy (headerLine, s);
			}
		else
			{
			if (nNumbersOnThisLine > 0)
				numParamLines++;
			}
		
		}
		
	/* Now, check some aspects of the .p file. */
	if (inSumpComment == YES)
		{
		MrBayesPrint ("%s   Unterminated comment in file %s\n", spacer, sumpParams.sumpFileName);
		goto errorExit;
		}
	if (numParamLines <= 0)
		{
		MrBayesPrint ("%s   No parameters were found in file %s\n", spacer, sumpParams.sumpFileName);
		goto errorExit;
		}
	if (sumpParams.sumpBurnIn > numParamLines)
		{
		MrBayesPrint ("%s   No parameters are sampled as the burnin exceeds the number of lines in last block\n", spacer);
		MrBayesPrint ("%s   Try setting burnin to a number less than %d\n", spacer, numParamLines);
		goto errorExit;
		}
		
	/* tell the user that everything is fine */
	MrBayesPrint ("%s   Found %d parameter lines in file \"%s\"\n", spacer, numParamLines, sumpParams.sumpFileName);
	if (sumpParams.sumpBurnIn > 0)
		MrBayesPrint ("%s   Of the %d lines, %d of them will be summarized (starting at line %d)\n", spacer, numParamLines, numParamLines - sumpParams.sumpBurnIn, lastNonDigitLine + sumpParams.sumpBurnIn + 1);
	else
		MrBayesPrint ("%s   All %d lines will be summarized (starting at line %d)\n", spacer, numParamLines, lastNonDigitLine+1);
	MrBayesPrint ("%s   (Only the last set of lines will be read, in case multiple\n", spacer);
	MrBayesPrint ("%s   parameter blocks are present in the same file.)\n", spacer);
	
	/* Calculate and check the number of columns and rows for the file */
	(void)fseek(fp, 0L, 0);	
	for (lineNum=0; lineNum<lastNonDigitLine+sumpParams.sumpBurnIn; lineNum++)
		fgets (s, longestLineLength, fp);
	inSumpComment = NO;
	nLines = 0;
	numRows = numColumns = 0;
	while (fgets (s, longestLineLength, fp) != NULL)
		{
		sumpTokenP = &s[0];
		allDigitLine = YES;
		lastTokenWasDash = NO;
		nNumbersOnThisLine = 0;
		do
			{
			GetSumpToken (&tokenType);
			if (IsSame("[", sumpToken) == SAME)
				inSumpComment = YES;
			if (IsSame("]", sumpToken) == SAME)
				inSumpComment = NO;
			if (inSumpComment == NO)
				{
				if (tokenType == NUMBER)
					{
					nNumbersOnThisLine++;
					lastTokenWasDash = NO;
					}
				else if (tokenType == DASH)
					{
					lastTokenWasDash = YES;
					}
				else if (tokenType != UNKNOWN_TOKEN_TYPE)
					{
					allDigitLine = NO;
					lastTokenWasDash = NO;
					}
				}
			} while (*sumpToken);
		lineNum++;
		if (allDigitLine == NO)
			{
			MrBayesPrint ("%s   Found a line with non-digit characters (line %d)\n", spacer, lineNum);
			goto errorExit;
			}
		else
			{
			if (nNumbersOnThisLine > 0)
				{
				nLines++;
				if (nLines == 1)
					firstNumCols = nNumbersOnThisLine;
				else
					{
					if (nNumbersOnThisLine != firstNumCols)
						{
						MrBayesPrint ("%s   Number of lines is not even (%d in first line and %d in %d line)\n", spacer, firstNumCols, nNumbersOnThisLine, lineNum);
						goto errorExit;
						}
					}
				}
			}
		}
	numRows = nLines;
	numColumns = firstNumCols;
	MrBayesPrint ("%s   %d rows and %d columns in each row\n", spacer, numRows, numColumns);
	
	/* allocate space to hold parameter information */
	if (numRows == 0 || numColumns == 0)
		{
		MrBayesPrint ("%s   The number of rows or columns is equal to zero\n", spacer);
		goto errorExit;
		}
	if (memAllocs[ALLOC_SUMPINFO] == YES)
		{
		MrBayesPrint ("%s   Sump string is already allocated\n", spacer);
		goto errorExit;
		}
	parameterValues = (MrBFlt *)malloc((size_t) (numRows * numColumns * sizeof(MrBFlt)));
	if (!parameterValues)
		{
		MrBayesPrint ("%s   Problem allocating parameterValues\n", spacer);
		goto errorExit;
		}
	memAllocs[ALLOC_SUMPINFO] = YES;
	for (i=0; i<numRows*numColumns; i++)
		parameterValues[i] = 0.0;
	/* Now we read the file for real. First, rewind file pointer to beginning of file... */
	(void)fseek(fp, 0L, 0);	
	
	/* ...and fast forward to beginning of last unburned parameter line. */
	for (lineNum=0; lineNum<lastNonDigitLine+sumpParams.sumpBurnIn; lineNum++)
		fgets (s, longestLineLength, fp);
		
	/* ...and parse file, line-by-line. We are only parsing lines that have digits that should be read. */
	inSumpComment = NO;
	numLinesToRead = numParamLines - sumpParams.sumpBurnIn;
	numLinesRead = j = 0;
	while (fgets (s, longestLineLength, fp) != NULL)
		{
		sumpTokenP = &s[0];
		allDigitLine = YES;
		lastTokenWasDash = NO;
		nNumbersOnThisLine = 0;
		do
			{
			GetSumpToken (&tokenType);
			if (IsSame("[", sumpToken) == SAME)
				inSumpComment = YES;
			if (IsSame("]", sumpToken) == SAME)
				inSumpComment = NO;
			if (inSumpComment == NO)
				{
				if (tokenType == NUMBER)
					{
					/* read the information from this line */
					if (j >= numRows * numColumns)
						{
						MrBayesPrint ("%s   Too many parameter values read in (%d)\n", spacer, j);
						goto errorExit;
						}
					sscanf (sumpToken, "%lf", &tempD);
					if (lastTokenWasDash == YES)
						tempD *= -1.0;
					parameterValues[numLinesRead * numColumns + nNumbersOnThisLine] = tempD;
					j++;
					nNumbersOnThisLine++;
					lastTokenWasDash = NO;
					}
				else if (tokenType == DASH)
					{
					lastTokenWasDash = YES;
					}
				else if (tokenType != UNKNOWN_TOKEN_TYPE)
					{
					/* we have a problem */
					MrBayesPrint ("%s   Found a line with non-digit characters (line %d)\n", spacer, lineNum);
					goto errorExit;
					}
				}
			} while (*sumpToken);
		lineNum++;
		if (nNumbersOnThisLine > 0)
			numLinesRead++;
		}
	
	/* tell user how many lines were successfully read */
	MrBayesPrint ("%s   Successfully read %d lines from last parameter block\n", spacer, numLinesRead);
	
	/* Check that at least one parameter line was read in. */
	if (numLinesRead <= 0)
		{
		MrBayesPrint ("%s   No parameters read in\n", spacer);
		goto errorExit;
		}
				
	/* separate header line into titles for each column */
	if (GetHeaders (headerLine, &nHeaders) == ERROR)
		goto errorExit;
		
	/* get length of longest header */
	longestHeader = 9; /* 9 is the length of the word "parameter" (for printing table) */
	for (i=0; i<nHeaders; i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			goto errorExit;
			}
		len = strlen(temp);
		if (len > longestHeader)
			longestHeader = len;
		}
		
	/* print x-y plot of log likelihood vs. generation */
	screenWidth = 60; /* don't change this without changing numY and meanY, declared above */
	screenHeigth = 15;
	whichIsX = whichIsY = -1;
	for (i=0; i<nHeaders; i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			goto errorExit;
			}
		len = strlen(temp);
		if (IsSame (temp, "Gen") == SAME)
			whichIsX = i;
		if (IsSame (temp, "lnL") == SAME)
			whichIsY = i;
		}		
	if (whichIsX >= 0 && whichIsX >= 0)
		{			
		minX = minY = 1000000000.0;
		maxX = maxY = -1000000000.0;
		for (i=0; i<numRows; i++)
			{
			x = parameterValues[i * numColumns + whichIsX];
			y = parameterValues[i * numColumns + whichIsY];
			if (x < minX)
				minX = x;
			if (y < minY)
				minY = y;
			if (x > maxX)
				maxX = x;
			if (y > maxY)
				maxY = y;
			}
		for (i=0; i<screenWidth; i++)
			{
			numY[i] = 0;
			meanY[i] = 0.0;
			}
		for (i=0; i<numRows; i++)
			{
			x = parameterValues[i * numColumns + whichIsX];
			y = parameterValues[i * numColumns + whichIsY];
			k = (int)(((x - minX) / (maxX - minX)) * screenWidth);
			if (k >= screenWidth)
				k = screenWidth - 1;
			meanY[k] += y;
			numY[k]++;
			}
		MrBayesPrint ("\n");
		MrBayesPrint ("%s   Below is a rough plot of the generation (x-axis) versus the log   \n", spacer);
		MrBayesPrint ("%s   probability of observing the data (y-axis).  You can use this     \n", spacer);
		MrBayesPrint ("%s   graph to determine what the burn-in for your analysis should be.  \n", spacer);
		MrBayesPrint ("%s   When the log probability starts to plateau you may be at station- \n", spacer);
		MrBayesPrint ("%s   arity. Sample trees and parameters after the log probability      \n", spacer);
		MrBayesPrint ("%s   plateaus. Of course, this is not a guarantee that you are at      \n", spacer);
		MrBayesPrint ("%s   stationarity. When possible, run multiple chains starting from    \n", spacer);
		MrBayesPrint ("%s   random trees; if the inferences you make for independent chains   \n", spacer);
		MrBayesPrint ("%s   are the same, this is reasonable evidence that the chains have    \n", spacer);
		MrBayesPrint ("%s   converged. Remember that the burn in is the number of samples     \n", spacer);
		MrBayesPrint ("%s   to discard. There are a total of ngen / samplefreq samples taken  \n", spacer);
		MrBayesPrint ("%s   during a MCMC analysis.                                           \n", spacer);
			
		MrBayesPrint ("\n   +");
		for (i=0; i<screenWidth; i++)
			MrBayesPrint ("-");
		MrBayesPrint ("+ %1.2lf\n", maxY);
		for (j=screenHeigth-1; j>=0; j--)
			{
			MrBayesPrint ("   |");
			for (i=0; i<screenWidth; i++)
				{
				if (numY[i] > 0)
					{
					if (meanY[i] / numY[i] > (((maxY - minY)/screenHeigth)*j)+minY && meanY[i] / numY[i] <= (((maxY - minY)/screenHeigth)*(j+1))+minY)
						MrBayesPrint ("*");
					else
						MrBayesPrint (" ");
					}
				else
					{
					MrBayesPrint (" ");
					}
				}
			MrBayesPrint ("|\n");
			}
		MrBayesPrint ("   +");
		for (i=0; i<screenWidth; i++)
			{
			if (i % (screenWidth/10) == 0 && i != 0)
				MrBayesPrint ("+");
			else
				MrBayesPrint ("-");
			}
		MrBayesPrint ("+ %1.2lf\n", minY);
		MrBayesPrint ("   ^");
		for (i=0; i<screenWidth; i++)
			MrBayesPrint (" ");
		MrBayesPrint ("^\n");
		MrBayesPrint ("   %1.0lf", minX);
		for (i=0; i<screenWidth; i++)
			MrBayesPrint (" ");
		MrBayesPrint ("%1.0lf\n\n", maxX);
		}
			
	/* sort parameter values for each column */
	if (SortParameters () == ERROR)
		goto errorExit;
		
	/* deal with lnL */
	for (i=0; i<nHeaders; i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			goto errorExit;
			}
		len = strlen(temp);
		if (IsSame (temp, "lnL") == SAME)
			{
			if (MeanVariance(i, &mean, &var, YES) == ERROR)
				{
				MrBayesPrint ("\n");
				MrBayesPrint ("%s   Estimated marginal likelihood = %1.2lf (arithmetic mean) *\n",  spacer, mean);
				MrBayesPrint ("%s                                 = %1.2lf (harmonic mean) *\n\n",  spacer, var);
				MrBayesPrint ("%s   * Note that these estimates may be unreliable because \n", spacer);
				MrBayesPrint ("%s     some extreme values were excluded\n\n", spacer);
				
				}
			else
				{
				MrBayesPrint ("\n");
				MrBayesPrint ("%s   Estimated marginal likelihood = %1.2lf (arithmetic mean)\n",  spacer, mean);
				MrBayesPrint ("%s                                 = %1.2lf (harmonic mean)\n\n",  spacer, var);
				}
			}
		else
			continue;
			
		}	
	/* Print parameter information to screen. */	
	MrBayesPrint ("%s   Model parameter summaries:\n\n", spacer);
	MrBayesPrint ("%s   (Based on a total of %d samples out of a total of %d samples recorded)\n\n", spacer, numParamLines - sumpParams.sumpBurnIn, numParamLines);
	MrBayesPrint ("%s            ", spacer);
	for (j=0; j<longestHeader - 9; j++)
		MrBayesPrint (" ");
	MrBayesPrint ("                           95%% Cred. Interval\n");
	MrBayesPrint ("%s            ", spacer);
	for (j=0; j<longestHeader - 9; j++)
		MrBayesPrint (" ");
	MrBayesPrint ("                          --------------------\n");
	MrBayesPrint ("%s   Parameter", spacer);
	for (j=0; j<longestHeader - 9; j++)
		MrBayesPrint (" ");
	MrBayesPrint ("    Mean      Variance      Lower       Upper      Median\n");
	MrBayesPrint ("%s   ---------", spacer);
	for (j=0; j<longestHeader - 9; j++)
		MrBayesPrint ("-");
	MrBayesPrint ("----------------------------------------------------------\n");
	for (i=0; i<nHeaders; i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			goto errorExit;
			}
		len = strlen(temp);
		if (IsSame (temp, "Gen") == SAME)
			continue;
		if (IsSame (temp, "Aamodel") == SAME || IsSame (temp, "Aamodel") == CONSISTENT_WITH)
			continue;
		if (IsSame (temp, "lnL") == SAME)
			continue;
		if (MeanVariance(i, &mean, &var, NO) == ERROR)
			{
			MrBayesPrint ("%s   Error getting mean and variance\n", spacer);
			goto errorExit;
			}
		lower = parameterValues[((int)(numRows * 0.025) * numColumns + i)];
		upper = parameterValues[((int)(numRows * 0.975) * numColumns + i)];
		MrBayesPrint ("%s   %s", spacer, temp);
		for (j=0; j<longestHeader - len; j++)
			MrBayesPrint (" ");
			
		sprintf (val, "%0.6lf", mean);
		j = 0;
		while (val[j] != '.')
			j++;
		if (j == 1)
			MrBayesPrint ("  ");
		else if (j == 2)
			MrBayesPrint (" ");
		j = 0;
		while (val[j] != '.')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint (".");
		j++;
		while (val[j] != '\0')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint ("  ");
		sprintf (val, "%0.6lf", var);
		j = 0;
		while (val[j] != '.')
			j++;
		if (j == 1)
			MrBayesPrint ("  ");
		else if (j == 2)
			MrBayesPrint (" ");
		j = 0;
		while (val[j] != '.')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint (".");
		j++;
		while (val[j] != '\0')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint ("  ");
		
		sprintf (val, "%0.6lf", lower);
		j = 0;
		while (val[j] != '.')
			j++;
		if (j == 1)
			MrBayesPrint ("  ");
		else if (j == 2)
			MrBayesPrint (" ");
		j = 0;
		while (val[j] != '.')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint (".");
		j++;
		while (val[j] != '\0')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint ("  ");
		sprintf (val, "%0.6lf", upper);
		j = 0;
		while (val[j] != '.')
			j++;
		if (j == 1)
			MrBayesPrint ("  ");
		else if (j == 2)
			MrBayesPrint (" ");
		j = 0;
		while (val[j] != '.')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint (".");
		j++;
		while (val[j] != '\0')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint ("  ");
			
		sprintf (val, "%0.6lf", parameterValues[((int)(numRows * 0.50) * numColumns + i)]);
		j = 0;
		while (val[j] != '.')
			j++;
		if (j == 1)
			MrBayesPrint ("  ");
		else if (j == 2)
			MrBayesPrint (" ");
		j = 0;
		while (val[j] != '.')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint (".");
		j++;
		while (val[j] != '\0')
			MrBayesPrint ("%c", val[j++]);
		MrBayesPrint ("  ");
		MrBayesPrint ("\n");
		}
	MrBayesPrint ("%s   ---------", spacer);
	for (j=0; j<longestHeader - 9; j++)
		MrBayesPrint ("-");
	MrBayesPrint ("----------------------------------------------------------\n");
	
	/* deal with amino acid model probabilities */
	for (i=0; i<nHeaders; i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			goto errorExit;
			}
		len = strlen(temp);
		if (IsSame (temp, "Aamodel") == SAME || IsSame (temp, "Aamodel") == CONSISTENT_WITH)
			{
			MrBayesPrint ("\n\n");
			if (IsSame (temp, "Aamodel") == SAME)
				MrBayesPrint ("%s   Amino acid model probabilities:\n\n", spacer);
			else
				{
				startReading = NO;
				j = k = 0;
				while (temp[j] != '\0')
					{
					if (startReading == YES && temp[j] != '}')
						parts[k++] = temp[j];
					if (temp[j] == '{')
						startReading = YES;
					else if (temp[j] == '}')
						startReading = NO;
					j++;
					}
				parts[k] = '\0';
				MrBayesPrint ("%s   Amino acid model probabilities for partition(s) %s:\n\n", spacer, parts);
				}
			for (j=0; j<20; j++)
				aaModelCounts[j] = 0;
			for (j=0; j<numRows; j++)
				aaModelCounts[(int)(parameterValues[j * numColumns + i])]++;
			MrBayesPrint ("%s              Posterior\n", spacer);
			MrBayesPrint ("%s   Model     Probability\n", spacer);
			MrBayesPrint ("%s   -------------------\n", spacer);
			MrBayesPrint ("%s   Poisson      %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_POISSON] / numRows);
			MrBayesPrint ("%s   Jones        %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_JONES]   / numRows);
			MrBayesPrint ("%s   Dayhoff      %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_DAY]     / numRows);
			MrBayesPrint ("%s   Mtrev        %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_MTREV]   / numRows);
			MrBayesPrint ("%s   Mtmam        %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_MTMAM]   / numRows);
			MrBayesPrint ("%s   Wag          %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_WAG]     / numRows);
			MrBayesPrint ("%s   Rtrev        %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_RTREV]   / numRows);
			MrBayesPrint ("%s   Cprev        %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_CPREV]   / numRows);
			MrBayesPrint ("%s   Vt           %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_VT]      / numRows);
			MrBayesPrint ("%s   Blossum      %1.3lf\n", spacer, (MrBFlt)aaModelCounts[AAMODEL_BLOSSUM] / numRows);
			MrBayesPrint ("%s   -------------------\n", spacer);
			}
		}
	/* free memory and file pointers */
	if (memAllocs[ALLOC_SUMPSTRING] == YES)
		{
		free (s);
		free (headerLine);
		free (headerNames);
		memAllocs[ALLOC_SUMPSTRING] = NO;
		}
	if (memAllocs[ALLOC_SUMPINFO] == YES)
		{
		free (parameterValues);
		memAllocs[ALLOC_SUMPINFO] = NO;
		}
	if (fp)
		fclose (fp);
	expecting = Expecting(COMMAND);
	
#	if defined (MPI_ENABLED)
		}
	ierror = MPI_Barrier (MPI_COMM_WORLD);
	if (ierror != MPI_SUCCESS)
		{
		MrBayesPrint ("%s   Problem at chain barrier\n", spacer);
		goto errorExit;
		}
#	endif
	return (NO_ERROR);
	
	errorExit:
		expecting = Expecting(COMMAND);
		if (memAllocs[ALLOC_SUMPSTRING] == YES)
			{
			if (s)
				free (s);
			if (headerLine)
				free (headerLine);
			if (headerNames)
				free (headerNames);
			memAllocs[ALLOC_SUMPSTRING] = NO;
			}
		if (memAllocs[ALLOC_SUMPINFO] == YES)
			{
			free (parameterValues);
			memAllocs[ALLOC_SUMPINFO] = NO;
			}
		if (fp != NULL)
			fclose (fp);
		strcpy (spacer, "");
		strcpy (sumpToken, "Sump");
		i = 0;
		if (FindValidCommand (sumpToken, &i) == ERROR)
			MrBayesPrint ("%s   Could not find sump\n", spacer);
		return (ERROR);	
	
}
int DoSumpParm (char *parmName, char *tkn)
{
	int			tempI;
	if (defMatrix == NO)
		{
		MrBayesPrint ("%s   A matrix must be specified before sumt can be used\n", spacer);
		return (ERROR);
		}
	if (expecting == Expecting(PARAMETER))
		{
		expecting = Expecting(EQUALSIGN);
		}
	else
		{
		if (!strcmp(parmName, "Xxxxxxxxxx"))
			{
			expecting  = Expecting(PARAMETER);
			expecting |= Expecting(SEMICOLON);
			}
		/* set Filename (sumpParams.sumpFileName) ***************************************************/
		else if (!strcmp(parmName, "Filename"))
			{
			if (expecting == Expecting(EQUALSIGN))
				expecting = Expecting(ALPHA);
			else if (expecting == Expecting(ALPHA))
				{
				strcpy (sumpParams.sumpFileName, tkn);
				MrBayesPrint ("%s   Setting sump filename to %s\n", spacer, sumpParams.sumpFileName);
				expecting = Expecting(PARAMETER) | Expecting(SEMICOLON);
				}
			else
				return (ERROR);
			}
		/* set Burnin (sumpParams.sumpBurnIn) *******************************************************/
		else if (!strcmp(parmName, "Burnin"))
			{
			if (expecting == Expecting(EQUALSIGN))
				expecting = Expecting(NUMBER);
			else if (expecting == Expecting(NUMBER))
				{
				sscanf (tkn, "%d", &tempI);
				sumpParams.sumpBurnIn = tempI;
				MrBayesPrint ("%s   Setting sump burnin to %d\n", spacer, sumpParams.sumpBurnIn);
				expecting = Expecting(PARAMETER) | Expecting(SEMICOLON);
				}
			else
				return (ERROR);
			}
		else
			return (ERROR);
		}
	return (NO_ERROR);
}
int GetHeaders (char *h, int *nHeaders)
{
	int			i, j, x, wasLastCharSpace;
	char		temp[100];
	wasLastCharSpace = YES;
	i = j = (*nHeaders) = 0;
	while (h[i] != '\0')
		{
		if (h[i] == ' ' || h[i] == '\t')
			{
			if (wasLastCharSpace == NO)
				{
				temp[j++] = '\0';
				if (AddToString (temp, headerNames, &x) == ERROR)
					{
					MrBayesPrint ("%s   Error adding header to list of headers \n", spacer);
					return ERROR;
					}
				(*nHeaders)++;
				}
			wasLastCharSpace = YES;
			j = 0;
			}
		else
			{
			temp[j++] = h[i];
			wasLastCharSpace = NO;
			}
		i++;
		}
		
#	if 0
	for (i=0; i<(*nHeaders); i++)
		{
		if (GetNameFromString (headerNames, temp, i+1) == ERROR)
			{
			MrBayesPrint ("%s   Error getting header names \n", spacer);
			return (ERROR);
			}
		printf ("%4d -> %s\n", i, temp);
		}
#	endif
		
	return (NO_ERROR);
	
}
void GetSumpToken (int *tokenType)
{
		
	int				allNumbers;
	register char	*temp;
	
	(*tokenType) = 0;
	temp = sumpToken;
	
	while (IsWhite(*sumpTokenP) == 1 || IsWhite(*sumpTokenP) == 2)
		{
		if (IsWhite(*sumpTokenP) == 2)
			{
			*tokenType = RETURNSYMBOL;
			foundNewLine = YES;
			/* MrBayesPrint ("RETURN\n"); */
			}
		++sumpTokenP;
		}
	
	*tokenType = UNKNOWN_TOKEN_TYPE;
	if (IsIn(*sumpTokenP,"="))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = EQUALSIGN;
		}
	else if (IsIn(*sumpTokenP,";"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = SEMICOLON;
		}
	else if (IsIn(*sumpTokenP,":"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = COLON;
		}
	else if (IsIn(*sumpTokenP,","))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = COMMA;
		}
	else if (IsIn(*sumpTokenP,"#"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = POUNDSIGN;
		}
	else if (IsIn(*sumpTokenP,"("))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = LEFTPAR;
		}
	else if (IsIn(*sumpTokenP,")"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = RIGHTPAR;
		}
	else if (IsIn(*sumpTokenP,"{"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = LEFTCURL;
		}
	else if (IsIn(*sumpTokenP,"}"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = RIGHTCURL;
		}
	else if (IsIn(*sumpTokenP,"["))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = LEFTCOMMENT;
		}
	else if (IsIn(*sumpTokenP,"]"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = RIGHTCOMMENT;
		}
	else if (IsIn(*sumpTokenP,"?"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = QUESTIONMARK;
		}
	else if (IsIn(*sumpTokenP,"-"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = DASH;
		}
	else if (IsIn(*sumpTokenP,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789."))
		{
		allNumbers = TRUE;
		if (!IsIn(*sumpTokenP,"0123456789."))
			allNumbers = FALSE;
		*temp++ = *sumpTokenP++;
		while(IsIn(*sumpTokenP,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789."))
			{
			if (!IsIn(*sumpTokenP,"0123456789."))
				allNumbers = FALSE;
			*temp++ = *sumpTokenP++;
			}
		if (allNumbers == TRUE)
			*tokenType = NUMBER;
		else
			*tokenType = ALPHA;
		}
	else if (IsIn(*sumpTokenP,"*"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = ASTERISK;
		}
	else if (IsIn(*sumpTokenP,"/"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = BACKSLASH;
		}
	else if (IsIn(*sumpTokenP,"'\\'"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = FORWARDSLASH;
		}
	else if (IsIn(*sumpTokenP,"!"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = EXCLAMATIONMARK;
		}
	else if (IsIn(*sumpTokenP,"%"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = PERCENT;
		}
	else if (IsIn(*sumpTokenP,"\""))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = QUOTATIONMARK;
		}
	else if (IsIn(*sumpTokenP,"&~+^$@|{}`><"))
		{
		*temp++ = *sumpTokenP++;
		*tokenType = WIERD;
		}
	*temp = '\0';
	
}
int MeanVariance(int col, MrBFlt *m, MrBFlt *v, int isLog)
{
	int				i, reliable;
	MrBFlt			a, aOld, s, x, y, scaler, n;
	reliable = YES;
	
	if (isLog == YES)
		{
		scaler = parameterValues[(numRows-1) * numColumns + col];
		a = aOld = n = 0.0;
		for (i=0; i<numRows; i++)
			{
			x = parameterValues[i * numColumns + col];
			y = x;
			y -= scaler;
			if (y < -100.0)
				{
				reliable = NO;
				continue;
				}
			else
				x = exp(y);
				
			if (n < 0.5)
				a = x;
			else
				{
				aOld = a;
				a = aOld + (x - aOld) / (n+1);
				}
			n += 1.0;
			}
		/* arithmetic mean */
		(*m) = log(a) + scaler;
		
		scaler = 0.0 - parameterValues[(numRows-1) * numColumns + col];
		a = aOld = n = 0.0;
		for (i=0; i<numRows; i++)
			{
			x = 0.0 - parameterValues[i * numColumns + col];
			y = x;
			y -= scaler;
			if (y < -100.0)
				{
				reliable = NO;
				continue;
				}
			else if (y > 40.0)
				{
				reliable = NO;
				continue;
				}
			else
				x = exp(y);
				
			if (n < 0.5)
				a = x;
			else
				{
				aOld = a;
				a = aOld + (x - aOld) / (n+1.0);
				}
			n += 1.0;
			}
		/* harmonic mean */
		(*v) = log(n) - (log(a) + scaler + log(n));
		}
	else
		{
		a = aOld = s = 0.0;
		for (i=0; i<numRows; i++)
			{
			x = parameterValues[i * numColumns + col];
			if (i == 0)
				{
				a = x;
				s = 0;
				}
			else
				{
				aOld = a;
				a = aOld + (x - aOld) / (i+1);
				s = s + (x - aOld) * (x - aOld);
				}
			}
		/* mean */
		(*m) = a;
		
		/* variance */
		if (numRows <= 1)
			(*v) = 0.0;
		else
			(*v) = s / (numRows - 1);
		}
		
	if (reliable == YES)
		return (NO_ERROR);
	else
		return (ERROR);
	
}
void Sort (MrBFlt *item, int count)
{
	Sort2 (item, 0, count-1);
}
void Sort2 (MrBFlt *item, int left, int right)
{
	register int	i, j;
	MrBFlt			x, y;
	i = left;
	j = right;
	x = item[(left+right)/2];
	do 
		{
		while (item[i] < x && i < right)
			i++;
		while (x < item[j] && j > left)
			j--;
		if (i <= j)
			{
			y = item[i];
			item[i] = item[j];
			item[j] = y;
			
			i++;
			j--;
			}
		} while (i <= j);
	if (left < j)
		Sort2 (item, left, j);
	if (i < right)
		Sort2 (item, i, right);
}
int SortParameters (void)
{
	int			i, j;
	MrBFlt		*tempParamValues;
	
	/* allocate information for sorting */
	tempParamValues = (MrBFlt *)malloc((size_t) (numRows * sizeof(MrBFlt)));
	if (!tempParamValues)
		{
		MrBayesPrint ("%s   Problem allocating tempParamValues\n", spacer);
		goto errorExit;
		}
		
	/* sort each column */
	for (j=0; j<numColumns; j++)
		{
		for (i=0; i<numRows; i++)
			tempParamValues[i] = 0.0;
		for (i=0; i<numRows; i++)
			tempParamValues[i] = parameterValues[i * numColumns + j];
		Sort (tempParamValues, numRows);
		for (i=0; i<numRows; i++)
			parameterValues[i * numColumns + j] = tempParamValues[i];
		}
#	if 0
	for (i=0; i<numRows; i++)
		{
		for (j=0; j<numColumns; j++)
			printf ("%2.3lf ", parameterValues[i * numColumns + j]);
		printf ("\n");
		}
#	endif
	/* free memory */
	free (tempParamValues);
	return (NO_ERROR);
	
	errorExit:
		if (tempParamValues)
			free (tempParamValues);
		return (ERROR);
}
